import math, uuid, re

from .mode import BaseMode, register_mode
from ..llms.gpt3 import GPT3

@register_mode('synth-chat')
class SynthChatMode(BaseMode):

  title = 'ChattyGPT'

  line_sep = '-----'

  def __init__(self, state={}):
    self.llm = GPT3()
    self.seed = ''

    self.human_name = 'Eden'
    self.authority_name = 'Boss'

    self.ai_name = 'Delphi AI'
    self.ai_bio = "Delphi AI is a cutting edge AGI created by Richard Feynman and Isaac Asimov to improve people's lives. She is incredibly knowledgable about everything, especially programming and math."

    self.pinned_summary = f'{self.human_name} demanded I give highly structured responses formatted in Markdown (lists like "1.", headers like "# title", code blocks like ```js etc)!'
    self.prologue = [
      {
        'source': 'server',
        'text': "Hi! I'm Delphi AI, a cutting edge AGI created by Richard Feynman and Isaac Asimov to improve people's lives.",
      },
    ]

    self.summary_header = f'# {self.ai_name}\'s notes on her conversation with {self.human_name}:'
    self.conversation_header = f'# Live Chat Between {self.human_name} and {self.ai_name}:'

    self.max_summaries = 8
    self.soft_max_depth = 18
    self.min_rollup_tokens = 200

    self.max_response_tokens = 750
    self.max_prompt_tokens = 4000

    self.soft_max_message_tokens = 150
    self.soft_max_prompt_tokens = min(2000, self.max_prompt_tokens)

    # Uncomment for easy prompt token pressure
    # self.max_prompt_tokens = 1500
    # self.soft_max_prompt_tokens = 1000

    self.summaries = []
    self.recent_conversation = self.build_messages(self.prologue)

    if state.get('initialized'):
      self.load(state)

  def load_profile(self, profile=None, profile_path=''):
    if not profile:
      with open(profile_path, "r") as profile_file:
        profile = json.load(profile_file)
    self.human_name = profile['human_name']
    self.authority_name = profile['authority_name']
    self.ai_name = profile['ai_name']
    self.ai_bio = profile['ai_bio']
    self.pinned_summary = profile['pinned_summary']
    self.prologue = profile['prologue']

  def save(self):
    return {
      'initialized': True,
      'summaries': self.summaries,
      'recent_conversation': self.recent_conversation
    }
    pass

  def load(self, state):
    self.summaries = state['summaries']
    self.recent_conversation = state['recent_conversation']

  def ask(self, query):
    client_message = self.add_message(text=query, source='client')
    server_message = self.build_message(source='server')

    response = self.continue_conversation()
    for data in response:
      server_message['text'] += data
      yield data

    server_message['text'] = server_message['text'].strip()
    self.add_message(message=server_message)

  def continue_conversation(self, stream=True):
    self.condense_conversation(self.max_response_tokens)

    conversation_prompt = self.format_conversation_prompt(self.recent_conversation)
    return self.llm.complete(
      conversation_prompt,
      max_length=self.max_response_tokens,
      stop=self.get_stops(),
      stream=stream
    )

  def build_message(self, text='', source=''):
    message = {
      'id': str(uuid.uuid4()),
      'text': text,
      'source': source,
    }
    return message

  def build_messages(self, messages):
    return [
      self.build_message(text=msg['text'], source=msg['source'])
      for msg in messages
    ]

  def add_message(self, message=None, text='', source=''):
    if not message:
      message = self.build_message(text=text, source=source)
    self.recent_conversation += [ message ]
    return message

  def delete_message(self, message={}, message_id=None):
    if not message_id:
      message_id = message['id']
    self.recent_conversation = [ m for m in self.recent_conversation if m['id'] != message_id ]

  def rollback(self):
    # TODO : do more stuff
    message_id = self.recent_conversation[-1]['id']
    self.delete_message(message_id=message_id)

  def shrink_messages(self):
    for (i, message) in enumerate(self.recent_conversation[:-8]):
      if not self.has_prompt_token_pressure():
        break
      if self.count_tokens(message['text']) < self.soft_max_message_tokens:
        continue

      # print('shrinking message')
      response = self.llm.complete(
        self.format_message_summary_prompt(message),
        stop=self.get_stops(),
      )
      message['text'] = f'{{ {response.strip()} }}'

  def condense_conversation(self, space_required):
    self.shrink_messages()

    while self.get_conversation_space() <= space_required or len(self.recent_conversation) > self.soft_max_depth:
      if not self.has_prompt_token_pressure():
        break

      # print('condensing conversation')

      chunk = []
      chunk_text = ''
      while self.count_tokens(chunk_text) < self.min_rollup_tokens and len(self.recent_conversation) > 2:
        chunk += self.recent_conversation[:1]
        self.recent_conversation = self.recent_conversation[1:]
        chunk_text = '\n'.join([ m['text'] for m in chunk ])

      if len(chunk) == 0:
        breakpoint()
        break

      summary_prompt = self.format_summary_prompt(chunk)
      response = self.llm.complete(
        summary_prompt,
        max_length=self.max_response_tokens,
        stop=self.get_stops(),
      )
      summary = f'{self.human_name} ' + response.strip()

      self.summaries = self.summaries[-(self.max_summaries - 1):] + [ summary ]

  def get_stops(self):
    return [ f'{self.human_name}>' ]

  def get_prompt_size(self):
    conversation_prompt = self.format_conversation_prompt(self.recent_conversation)
    return self.count_tokens(conversation_prompt)


  def get_conversation_space(self):
    return self.max_prompt_tokens - self.get_prompt_size() - self.max_response_tokens

  def has_prompt_token_pressure(self):
    return self.count_tokens(self.format_conversation_prompt(self.recent_conversation)) > self.soft_max_prompt_tokens;

  def get_notes(self):
    return [ f'- {s}' for s in self.summaries ]


  def format_message_summary_prompt(self, message, whitespace='\n'):
    return whitespace.join([
      f'Purpose: You are a genius AGI assistant named {self.ai_name} helping your friend {self.human_name} over chat.',
      self.format_message(message),
      f'{self.authority_name}> Briefly summarize the previous message from {"you" if message["source"] == "server" else self.human_name}, written as if you were {self.ai_name if message["source"] == "server" else self.human_name}. Should be short and dense but informative. Use the past tense.',
      f'{self.ai_name}>',
    ])

  def format_summary_prompt(self, messages, whitespace='\n'):
    return whitespace.join([
      self.ai_bio,
      self.conversation_header,
      self.format_messages(messages),
      f'{self.authority_name}> Give me your personal notes on the conversation. Should be terse but informative.',
      f'{self.ai_name}> {self.human_name}'
    ])

  def format_conversation_prompt(self, messages, whitespace='\n'):
    seed = self.seed.replace
    lines = [
      self.ai_bio,
      self.summary_header,
      f'- {self.pinned_summary}',
      *self.get_notes(),
      self.format_seed(),
      self.conversation_header,
      self.format_messages(messages),
      f'{self.ai_name}>'
    ]
    return whitespace.join(lines)

  def count_tokens(self, text):
    return self.llm.count_tokens(text)

  def format_seed(self):
    if not self.seed:
      return ''
    seed = re.sub(self.title, self.ai_name, self.seed, flags=re.IGNORECASE)
    seed = re.sub('{you}', self.ai_name, seed, flags=re.IGNORECASE)
    seed = re.sub('{me}', self.human_name, seed, flags=re.IGNORECASE)
    return f'# NOTE: Instructions For {self.ai_name} From {self.authority_name}: {seed}'
    # return f'Boss: {seed}'

  def format_messages(self, messages, wrap=True):
    lines = f'\n{self.line_sep}\n'.join(
      [ self.format_message(m) for m in messages ]
    )
    if wrap:
      lines = '\n'.join([ self.line_sep, lines, self.line_sep ])
    return lines

  def format_message(self, message):
    if message['source'] == 'server':
      return f'{self.ai_name}> {message["text"]}'
    else:
      return f'{self.human_name}> {message["text"]}'

  def print(self):
    print(self.format_conversation_prompt(self.recent_conversation))

  def stats(self):
    history_size = len(self.recent_conversation)
    headroom = 100 - round(100 * self.get_prompt_size() / self.max_prompt_tokens)
    return f'( depth={history_size}, free={headroom}% )'

  def capacity(self):
    return round(100 * self.get_prompt_size() / self.max_prompt_tokens)



import os, re, pyperclip

from .utils import printer, open_in_editor
from .publish import PublishGPT

class CommandsClass:

  def __init__(self):
    pass

  def exec(self, cli, text):
    text = text.strip()

    if text == '':
      pass

    elif text == '.clear' or text == '.cl':
      self.clear_screen()

    elif text == '.editor' or text == '.e':
      print()
      cli.session.layout.current_buffer.text = ""
      cli.session.layout.current_buffer.open_in_editor()

    elif text == '.config':
      pass

    elif text == '.exit':
      print()
      raise KeyboardInterrupt

    elif text == '.copy' or text == '.cp':
      self.print_command_banner('.copy')
      code_block = self.extract_recent_code_block(cli.thread["history"])
      if code_block == None:
        print("No code block found")
      else:
        success = self.copy_to_clipboard(code_block)
        if success:
          print('Copied the first code block in the most recent response to the clipboard')
        else:
          print("Failed to copy to the clipboard")
      print()

    elif text == '.publish' or text == '.pub':
      self.print_command_banner('.publish')
      publisher = PublishGPT(cli.thread)
      url = publisher.publish()
      print(url)
      print()

    elif text == ('.thread') or text == ('.name'):
      self.print_command_banner('.thread')
      print(f"Thread name is \"{cli.thread['id']}\"")
      print()

    elif text.startswith('.seed') or text.startswith('.s ') or text == '.s':
      self.print_command_banner('.seed')
      cli.mode.set_seed(' '.join(text.split(' ')[1:]).strip())
      print('Seed updated\n')

    elif text.startswith('.rename'):
      self.print_command_banner('.rename')

      parts = text.split(' ')
      if len(parts) == 1:
        print(f"Thread name is \"{cli.thread['id']}\"")
      elif len(parts) == 2:
        cli.thread['id'] = parts[1]
        print(f"Thread renamed to \"{cli.thread['id']}\"")
      else:
        print("Invalid thread name")

      print()

    elif text == '.reset' or text == '.rs':
      cli.reset()
      self.clear_screen()

      self.print_command_banner('.reset', newline=False)
      print(f"@{cli.thread['id']} has been reset")
      print()

    elif text == '.redo' or text == '.retry' or text == '.r':
      self.print_command_banner('.redo')
      history = cli.thread['history']
      if len(history) < 2:
        print('No message to redo\n')
        return (text, True)

      cli.mode.rollback()
      cli.mode.rollback()
      last_message = history[-2]['text']
      cli.thread['history'] = history[:-2]

      return (last_message, False)

    elif text == '.help' or text == 'help':
      self.print_command_banner('.help')
      self.print_help()
      print()

    elif text == '.threads':
      self.print_command_banner('.threads')
      for thread in cli.config.list_threads():
        print(thread)

    elif text == '.debug' or text == '.db':
      cli.mode.debug()

    elif text == '.print' or text == '.pp':
      self.print_command_banner('.print')
      cli.mode.print()
      print()

    elif text[0] == '.':
      self.print_command_banner('.error')
      print(f"Invalid command. Use .help to list valid commands.\n")

    else:
      return (text, False)

    return (text, True)

  def print_command_banner(self, name, newline=True):
    if newline:
      print()
    printer.print_banner(
      bg_color='dark_red',
      text=f'[ {name} ] =>',
    )

  def print_help(self):
    printer.print_markdown("""
## Commands
**.clear** or .cl: Clears the terminal screen.
**.copy** or .cp: Copies the first code block in the most recent response to the clipboard.
**.debug** or .db: Triggers a breakpoint in the mode backend
**.editor** or .e: Opens the current message in the text editor specified by $EDITOR.
**.exit**: Closes the REPL.
**.help** or help: Prints a list of available commands and a brief description of each.
**.print** or .pp: Prints the internal prompt representation
**.publish** or .pub: Publishes the current thread online and prints the URL.
**.retry**: or .r: Resubmits the most recent successful response.
**.rename**: Renames the current thread. The new name must be provided as an argument to this command, e.g. ".rename my_new_name". The old thread isn't deleted.
**.reset** or .rs: Resets the history of the thread.
**.seed** or .s: The the text following .seed will be set as the conversation seed. Set no text to clear the seed.
**.thread** or .name: Prints the current thread name.
**.threads**: Lists all threads.
## Shortcuts
**Enter**: Submits the current message.
**Tab**: Adds a new line.
**C+c**: Closes the REPL.
**C+d**: Closes the REPL.
**C+r**: Search message history.
**C+x-C+e**: Opens the current message in the text editor specified by $EDITOR.
## Tips:
- To shape your conversation, consider setting a conversation seed using the `.seed DESCRIPTION` command. Some examples:
    - `.seed You must use Markdown headers on every message`
    - `.seed You must contantly use exclamation marks`
    - `.seed You must speak in rhymes`

  Seeds work best when set early in a conversation and are best phrased as commands.
- If you don't like the most recent response, retry it with `.retry`
- If the thread has gone completely off the rails, reset it with `.reset`
 - Many commands have shorter aliases e.g. `.seed` has `.s`. See the command list for more info.
""")

  def extract_recent_code_block(self, history):
    responses = [ entry["text"] for entry in history if entry["type"] == 'gpt' ]
    if len(responses) == 0:
      return None
    last_response = responses[-1]

    # This regex is too naive but is easy
    code_block_pattern = re.compile(r"```[a-zA-Z0-9]*?\n(.*?)```", re.DOTALL)
    code_blocks = code_block_pattern.findall(last_response)

    if len(code_blocks) == 0:
      return None

    return code_blocks[0]

  def copy_to_clipboard(self, text):
    try:
      pyperclip.copy(text)
      return True
    except Exception as e:
      return False

  def clear_screen(self):
    os.system('cls' if os.name == 'nt' else 'clear')


Commands = CommandsClass()



import click, os, datetime
from os.path import abspath

from .repl import REPL
from .config import Config
from .modes import list_modes

def load_file(file_path):
  with open(file_path) as f:
    text = f.read()
    return f"Provide a brief description of the following text:\n\n[ Loaded from {file_path} ]\n```\n{text}\n```"

@click.command(help='Usually just `gpt-repl @thread-name`')
@click.argument("cmd", type=str, required=False)
@click.option('-m', '--mode', default='synth-chat', help="REPL mode to load")
@click.option('-t', '--thread', default=None, help="Thread name to open")
@click.option('-f', '--file', type=click.Path(exists=True), help="Path to text file to preload")
@click.option('-l', '--list', is_flag=True, help="List all threads")
def run(cmd, mode, thread, file, list):
  config_path = os.getenv('GPT_REPL_CONFIG_PATH', '~/.config/gpt_repl')

  if str(cmd) == 'help':
    ctx = click.get_current_context()
    click.echo(ctx.get_help())
    return

  if list or cmd == 'threads':
    for thread in get_threads(config_path):
      print(thread)
    return

  if str(cmd)[0] == '@':
    thread = cmd[1:]

  if str(cmd) == '-':
    threads = get_threads(config_path)
    thread = threads[0]

  autofills = []

  if file is not None:
    file_path = abspath(file)
    autofills.append(load_file(file_path))

  if thread == None:
    threads = get_threads(config_path)
    i = len(threads) + 1
    while True:
      thread = f'thread_{i}'
      i += 1
      if thread not in threads:
        break

  repl = REPL(
    mode_name=mode,
    thread_id=thread,
    config_path=config_path,
    autofills=autofills,
  )

  if cmd == 'modes':
    for mode in list_modes():
      print(mode)
    return

  repl.run()

def get_threads(config_path):
  config = Config(config_path)
  return config.list_threads()

if __name__ == '__main__':
  run()



import sys

from prompt_toolkit import PromptSession
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.history import FileHistory
from prompt_toolkit.application.current import get_app
from prompt_toolkit import filters as Filters

from .utils import peek, printer, Loader
from .config import Config

# from .modes.synth_chat import SynthChatMode as ReplMode
from .modes import get_mode
from .commands import Commands

class REPL:

  def __init__(
      self,
      config_path="~/.config/gpt_repl",
      thread_id=None,
      mode_name = None,
      autofills=[],
    ):

    self.config = Config(config_path)
    self.config.load_plugins()

    self.thread = self.config.load_thread(thread_id)
    self.mode_name = self.thread.get('mode') or mode_name or 'synth-chat'
    self.autofills = autofills

    self.first_run = True

  def run(self):
    self.warmup()
    self.load_mode(self.mode_name)

    print("Enter 'help' for a list of commands. Use Enter to submit and Tab to start a new line.\n")
    self.replay_thread()
    self.start_prompt_session()

    while True:
      try:
        self.print_you_banner(len(self.thread["history"]) + 1)
        printer.pad_down(3)

        default = ''
        if len(self.autofills) > 0:
          default = self.autofills.pop(0)

        text = self.prompt(default=default)
        printer.print_markdown(text)

        ( text, stop ) = Commands.exec(self, text)
        if stop:
          continue

        print('')
        self.print_gpt_banner(len(self.thread["history"]) + 2, stats=self.mode.stats())
        printer.pad_down(3)

        try:
          answer = self.ask(text)
        except (KeyboardInterrupt, EOFError):
          self.mode.rollback()
          print('Request canceled\n')
          continue

        printer.print_markdown(answer)
        print('')

        self.thread["history"] += [
          { "type": "you", "text": text },
          { "type": "gpt", "text": answer }
        ]

        self.save_thread()
        self.first_run = False
      except (KeyboardInterrupt, EOFError):
        self.save_thread()
        printer.print_thread_closed(self.thread['id'])
        break

      except Exception as e:
        breakpoint()
        printer.print_thread_closed(self.thread['id'])

        printer.exception(e)
        sys.exit(1)

  def start_prompt_session(self):
    self.session = PromptSession(
      erase_when_done=True,
      history=FileHistory(self.config.prompt_history_path),
    )
    self.kb = KeyBindings()

    @Filters.Condition
    def is_not_searching():
      return not get_app().layout.is_searching

    @self.kb.add('escape', 'enter', filter=is_not_searching)
    def _(event):
      event.current_buffer.insert_text('\n')
    @self.kb.add("tab", filter=is_not_searching)
    def _(event):
      prefix = event.current_buffer.document.leading_whitespace_in_current_line
      event.current_buffer.insert_text('\n' + prefix)
    @self.kb.add('enter', filter=is_not_searching)
    def _(event):
      if len(event.current_buffer.text.strip()) > 0:
        event.current_buffer.validate_and_handle()
      else:
        event.current_buffer.insert_text('\n')

  def prompt(self, default=''):
    seed = self.mode.get_seed()
    if seed:
      bottom_toolbar = f'seed={seed}'
    else:
      bottom_toolbar = 'No conversation seed set'

    text = self.session.prompt(
      '',
      multiline=True,
      key_bindings=self.kb,
      enable_open_in_editor=True,
      tempfile_suffix='.md',
      default=default,
      bottom_toolbar=bottom_toolbar,
    )
    return text.strip()

  def print_you_banner(self, count):
    printer.print_banner(
      bg_color='rgb(0,95,135)',
      text=' You:',
      prefix=f' {count} ',
      suffix=f' @{self.thread["id"]} [ {self.mode_name} ]'
    )

  def print_gpt_banner(self, count, stats=''):
    printer.print_banner(
      bg_color='spring_green4',
      text=f' {self.mode.title}:',
      prefix=f' {count} ',
      suffix=stats
    )

  def save_thread(self):
    if self.thread['id'] == None:
      return
    self.config.save_thread(self.thread['id'], {
      "id": self.thread['id'],
      "mode": self.mode_name,
      "history": self.thread["history"],
      'state': self.mode.save(),
    })

  def load_mode(self, mode_name):
    self.thread['mode'] = mode_name
    self.mode = get_mode(mode_name)(state=self.thread['state'])

  def reset(self):
    current_seed = self.mode.get_seed() if self.mode else ''

    self.thread = self.config.get_empty_thread(self.thread['id'])

    self.load_mode(self.mode_name)
    self.mode.set_seed(current_seed)

    self.save_thread()

  def replay_thread(self):
    for i, entry in enumerate(self.thread["history"]):
      if entry["type"] == "you":
        self.print_you_banner(i + 1)
      elif entry["type"] == "gpt":
        self.print_gpt_banner(i + 1)
      printer.print_markdown(entry["text"])
      print("")

  def warmup(self):
    messages = [ entry["text"] for entry in self.thread["history"] ] + self.autofills
    if any([ '```' in m for m in messages ]):
      with printer.print_thread_loading(self.thread['id']):
        printer.preload()

  def ask(self, text):
    delay = 0.25 if self.first_run else 1.5
    with Loader(show_timer=True, delay=delay) as spinner:
      gen = iter(self.mode.ask(text))
      response = peek(gen)[0]

    answer = ''
    with printer.live(transient=True) as screen:
      for data in response:
        answer += data
        markdown = printer.to_markdown(answer.lstrip() + '█', code_theme='default')
        display_text = markdown.to_text() + '\n\n\n'
        screen.update(display_text)

    return answer.strip()



import os, json, re, pdb, importlib.util
from os import path, listdir
from .utils import printer

class Config:

  def __init__(self, config_dir_path):
    self.config_dir_path = path.abspath(path.expanduser(config_dir_path))
    self.config_file_path = path.join(self.config_dir_path, "config.json")
    self.prompt_history_path = path.join(self.config_dir_path, "prompt_history")
    self.threads_dir_path = path.join(self.config_dir_path, "threads")
    self.plugins_dir_path = path.join(self.config_dir_path, "plugins")

    if not path.exists(self.config_dir_path):
      os.makedirs(self.config_dir_path)

    if not path.exists(self.threads_dir_path):
      os.mkdir(self.threads_dir_path)

    if not path.exists(self.plugins_dir_path):
      os.mkdir(self.plugins_dir_path)

    self.reload()

  def save(self):
      with open(self.config_file_path, "w") as config_file:
        json.dump(self.config or {}, config_file, indent=2)

  def reload(self):
    self.config = {}

    if not path.exists(self.config_file_path):
      self.save()

    with open(self.config_file_path, "r") as config_file:
      self.config = json.load(config_file)

    return self.config

  def set_auth(self, email, password):
    self.config["email"] = email
    self.config["password"] = password
    self.save()

  def get_auth(self):
    email = self.config.get("email", None)
    password = self.config.get("password", None)
    return (email, password)

  def normalize_thread_id(self, thread_id):
    if thread_id == None:
      return None
    thread_id = thread_id.replace('@', '').strip()
    return re.sub(r'[^_a-zA-Z0-9\.]+', '_', thread_id)

  def get_thread_file_path(self, thread_id):
    thread_id = self.normalize_thread_id(thread_id)
    thread_file_path = path.join(self.threads_dir_path, f"{thread_id}.json")
    return thread_file_path

  def load_thread(self, thread_id):
    thread_id = self.normalize_thread_id(thread_id)

    if thread_id == None:
      return self.get_empty_thread(None)

    thread_file_path = self.get_thread_file_path(thread_id)
    if not path.exists(thread_file_path):
      return self.get_empty_thread(thread_id)

    with open(thread_file_path, "r") as thread_file:
      thread_data = json.load(thread_file)

    if not thread_data['state']:
      thread_data['state'] = {}

    return thread_data

  def save_thread(self, thread_id, thread_data):
    thread_id = self.normalize_thread_id(thread_id)
    thread_file_path = self.get_thread_file_path(thread_id)
    with open(thread_file_path, "w") as thread_file:
      json.dump(thread_data, thread_file, indent=2)

  def get_empty_thread(self, thread_id):
    thread_id = self.normalize_thread_id(thread_id)
    return {
      "id": thread_id,
      "mode": "",
      'state': {},
      "history": [],
    }

  # thanks gpt
  def list_threads(self):
    threads = []
    for thread_file in listdir(self.threads_dir_path):
      if thread_file.endswith(".json"):
        thread_id = thread_file[:-5]  # remove the ".json" extension
        thread_id = thread_id.replace("_", "-")  # replace underscores with dashes
        thread_path = path.join(self.threads_dir_path, thread_file)
        modification_time = path.getmtime(thread_path)
        threads.append((modification_time, thread_id))

    # sort threads by modification time in descending order
    threads.sort(reverse=True)

    return [thread_id for _, thread_id in threads]

  def load_plugins(self):
    files = os.listdir(self.plugins_dir_path)
    for file in files:
      file_path = path.join(self.plugins_dir_path, file)
      if not file_path.endswith('.py'):
        continue
      try:
        spec = importlib.util.spec_from_file_location('plugins', file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
      except Exception as e:
        printer.exception(e)
        continue
